@@ -1,14 +1,18 @@
-from collections import OrderedDict, Mapping
+from collections import OrderedDict, Mapping, Iterable
 from typing import Any
 from urllib import parse
 
+from django.conf import settings
+from django.contrib.auth import get_user_model
+from django.contrib.auth.models import AbstractUser
 from django.core.exceptions import ImproperlyConfigured
 from django.core.exceptions import ValidationError as DjangoValidationError
 from django.core.urlresolvers import get_resolver, resolve, get_script_prefix, Resolver404
+from django.db.models import QuerySet
 from django.utils.datastructures import MultiValueDictKeyError
 from django.utils.encoding import uri_to_iri
 from rest_framework.exceptions import ValidationError
-from rest_framework.fields import SkipField, empty
+from rest_framework.fields import SkipField, empty, ReadOnlyField
 from rest_framework.fields import get_error_detail, set_value
 from rest_framework.relations import HyperlinkedRelatedField, ManyRelatedField, MANY_RELATION_KWARGS, Hyperlink
 from rest_framework.serializers import HyperlinkedModelSerializer, ListSerializer, ModelSerializer
@@ -19,16 +23,23 @@ from rest_framework.utils.serializer_helpers import ReturnDict
 
 from djangoldp.fields import LDPUrlField, IdURLField
 from djangoldp.models import Model
+from djangoldp.permissions import LDPPermissions
 
 
 class LDListMixin:
+    '''A Mixin used by the custom Serializers in this file'''
     child_attr = 'child'
 
+    # converts primitive data representation to the representation used within our application
     def to_internal_value(self, data):
         try:
+            # if this is a container, the data will be stored in ldp:contains
             data = data['ldp:contains']
         except (TypeError, KeyError):
             pass
+
+        if len(data) == 0:
+            return []
         if isinstance(data, dict):
             data = [data]
         if isinstance(data, str) and data.startswith("http"):
@@ -36,11 +47,47 @@ class LDListMixin:
 
         return [getattr(self, self.child_attr).to_internal_value(item) for item in data]
 
+    # converts internal representation to primitive data representation
     def to_representation(self, value):
+        '''
+        Permission on container :
+         - Can Add if add permission on contained object's type
+         - Can view the container is view permission on container model : container obj are filtered by view permission
+        '''
+        try:
+            child_model = getattr(self, self.child_attr).Meta.model
+        except AttributeError:
+            child_model = value.model
+
+        parent_model = None
+
+        if isinstance(value, QuerySet):
+            value = list(value)
+
+        if not isinstance(value, Iterable):
+            filtered_values = value
+            container_permissions = Model.get_permissions(child_model, self.context['request'].user, ['view', 'add'])
+        else:
+            # this is a container. Parent model is the containing object, child the model contained
+            try:
+                parent_model = Model.resolve_parent(self.context['request'].path)
+            except:
+                parent_model = child_model
+
+            # remove objects from the list which I don't have permission to view
+            filtered_values = list(
+                filter(lambda v: Model.get_permission_classes(v, [LDPPermissions])[0]().has_object_permission(
+                    self.context['request'], self.context['view'], v), value))
+            container_permissions = Model.get_permissions(child_model, self.context['request'].user, ['add'])
+            container_permissions.extend(
+                Model.get_permissions(parent_model, self.context['request'].user,
+                                      ['view']))
+        if not self.id.startswith('http'):
+            self.id = '{}{}{}'.format(settings.BASE_URL, Model.resource(parent_model), self.id)
         return {'@id': self.id,
                 '@type': 'ldp:Container',
-                'ldp:contains': super().to_representation(value),
-                'permissions': Model.get_permissions(value.model, self.context['request'].user, ['view', 'add'])
+                'ldp:contains': super().to_representation(filtered_values),
+                'permissions': container_permissions
                 }
 
     def get_attribute(self, instance):
@@ -150,7 +197,10 @@ class JsonLdField(HyperlinkedRelatedField):
 class JsonLdRelatedField(JsonLdField):
     def to_representation(self, value):
         try:
-            return {'@id': super().to_representation(value)}
+            if Model.is_external(value):
+                return {'@id': value.urlid}
+            else:
+                return {'@id': super().to_representation(value)}
         except ImproperlyConfigured:
             return value.pk
 
@@ -189,10 +239,19 @@ class JsonLdIdentityField(JsonLdField):
 
     def to_representation(self, value: Any) -> Any:
         try:
-            return Hyperlink(value.webid(), value)
+            if isinstance(value, str):
+                return Hyperlink(value, value)
+            else:
+                return Hyperlink(value.webid(), value)
         except AttributeError:
             return super().to_representation(value)
 
+    def get_attribute(self, instance):
+        if Model.is_external(instance):
+            return instance.urlid
+        else:
+            return super().get_attribute(instance)
+
 
 class LDPSerializer(HyperlinkedModelSerializer):
     url_field_name = "@id"
@@ -200,16 +259,12 @@ class LDPSerializer(HyperlinkedModelSerializer):
     serializer_url_field = JsonLdIdentityField
     ModelSerializer.serializer_field_mapping[LDPUrlField] = IdURLField
 
-    @property
-    def data(self):
-        return super().data
-
     def get_default_field_names(self, declared_fields, model_info):
         try:
             fields = list(self.Meta.model._meta.serializer_fields)
         except AttributeError:
             fields = super().get_default_field_names(declared_fields, model_info)
-        if 'request' in self._context and not self._context['request']._request.method == 'GET':
+        if 'request' in self._context and not (self._context['request']._request.method == 'GET' or self._context['request']._request.method == 'OPTIONS'):
             try:
                 fields.remove(self.Meta.model._meta.auto_author)
             except ValueError:
@@ -220,7 +275,14 @@ class LDPSerializer(HyperlinkedModelSerializer):
 
     def to_representation(self, obj):
         data = super().to_representation(obj)
-
+        slug_field = Model.slug_field(obj)
+        for field in data:
+            if isinstance(data[field], dict) and '@id' in data[field]:
+                data[field]['@id'] = data[field]['@id'].format(Model.container_id(obj), str(getattr(obj, slug_field)))
+        if 'urlid' in data and data['urlid'] is not None:
+            data['@id'] = data.pop('urlid')['@id']
+        if not '@id' in data:
+            data['@id'] = '{}{}'.format(settings.SITE_URL, Model.resource(obj))
         rdf_type = Model.get_meta(obj, 'rdf_type', None)
         rdf_context = Model.get_meta(obj, 'rdf_context', None)
         if rdf_type is not None:
@@ -233,10 +295,53 @@ class LDPSerializer(HyperlinkedModelSerializer):
         return data
 
     def build_field(self, field_name, info, model_class, nested_depth):
-        nested_depth = self.compute_depth(nested_depth, model_class)
-
         return super().build_field(field_name, info, model_class, nested_depth)
 
+    def build_property_field(self, field_name, model_class):
+        class JSonLDPropertyField(ReadOnlyField):
+            def to_representation(self, instance):
+                from djangoldp.views import LDPViewSet
+
+                if isinstance(instance, QuerySet) or isinstance(instance, Model):
+                    try:
+                        model_class = instance.model
+                    except:
+                        model_class = instance.__class__
+                    serializer_generator = LDPViewSet(model=model_class,
+                                                      lookup_field=Model.get_meta(model_class, 'lookup_field', 'pk'),
+                                                      permission_classes=Model.get_meta(model_class,
+                                                                                        'permission_classes',
+                                                                                        [LDPPermissions]),
+                                                      fields=Model.get_meta(model_class, 'serializer_fields', []),
+                                                      nested_fields=Model.get_meta(model_class, 'nested_fields', []))
+                    parent_depth = max(getattr(self.parent.Meta, "depth", 0) - 1, 0)
+                    serializer_generator.depth = parent_depth
+                    serializer = serializer_generator.build_read_serializer()(context=self.parent.context)
+                    if parent_depth is 0:
+                        serializer.Meta.fields = ["@id"]
+
+                    if isinstance(instance, QuerySet):
+                        data = list(instance)
+
+                        return {'@id': '{}{}{}/'.format(settings.SITE_URL, '{}{}/', self.source),
+                                '@type': 'ldp:Container',
+                                'ldp:contains': [serializer.to_representation(item) if item is not None else None for
+                                                 item
+                                                 in data],
+                                'permissions': Model.get_permissions(self.parent.Meta.model,
+                                                                     self.context['request'].user,
+                                                                     ['view', 'add'])
+                                }
+                    else:
+                        return serializer.to_representation(instance)
+                else:
+                    return instance
+
+        field_class = JSonLDPropertyField
+        field_kwargs = {}
+
+        return field_class, field_kwargs
+
     def build_standard_field(self, field_name, model_field):
         class JSonLDStandardField:
             parent_view_name = None
@@ -246,11 +351,13 @@ class LDPSerializer(HyperlinkedModelSerializer):
                 super().__init__(**kwargs)
 
             def get_value(self, dictionary):
+                if self.field_name == 'urlid':
+                    self.field_name = '@id'
                 try:
                     object_list = dictionary["@graph"]
                     if self.parent.instance is None:
                         obj = next(filter(
-                            lambda o: not hasattr(o, self.parent.url_field_name) or "./" in o[self.url_field_name],
+                            lambda o: not self.parent.url_field_name in o or "./" in o[self.parent.url_field_name],
                             object_list))
                         value = super().get_value(obj)
                     else:
@@ -261,6 +368,9 @@ class LDPSerializer(HyperlinkedModelSerializer):
                 except KeyError:
                     value = super().get_value(dictionary)
 
+                if self.field_name == '@id' and value == './':
+                    self.field_name = 'urlid'
+                    return None
                 return self.manage_empty(value)
 
             def manage_empty(self, value):
@@ -279,7 +389,6 @@ class LDPSerializer(HyperlinkedModelSerializer):
         return type(field_class.__name__ + 'Valued', (JSonLDStandardField, field_class), {}), field_kwargs
 
     def build_nested_field(self, field_name, relation_info, nested_depth):
-        nested_depth = self.compute_depth(nested_depth, self.Meta.model)
 
         class NestedLDPSerializer(self.__class__):
 
@@ -305,6 +414,7 @@ class LDPSerializer(HyperlinkedModelSerializer):
 
                     ret = OrderedDict()
                     errors = OrderedDict()
+
                     fields = list(filter(lambda x: x.field_name in data, self._writable_fields))
 
                     for field in fields:
@@ -340,35 +450,41 @@ class LDPSerializer(HyperlinkedModelSerializer):
                         slug_field = Model.slug_field(self.__class__.Meta.model)
                         ret[slug_field] = match.kwargs[slug_field]
                     except Resolver404:
-                        pass
+                        if 'urlid' in data:
+                            ret['urlid'] = data['urlid']
 
-                    return ret
                 else:
-                    return super().to_internal_value(data)
+                    ret = super().to_internal_value(data)
+
+                if self.url_field_name in data and not 'urlid' in data and data[self.url_field_name].startswith('http'):
+                    ret['urlid'] = data[self.url_field_name]
+
+                return ret
 
         kwargs = get_nested_relation_kwargs(relation_info)
         kwargs['read_only'] = False
         kwargs['required'] = False
         return NestedLDPSerializer, kwargs
 
-    @classmethod
-    def compute_depth(cls, depth, model_class, name='depth'):
-        try:
-            model_depth = getattr(model_class._meta, 'depth', getattr(model_class.Meta, 'depth', 10))
-            depth = min(depth, int(model_depth))
-        except AttributeError:
-            depth = min(depth, int(getattr(model_class._meta, 'depth', 1)))
-
-        return depth
-
     @classmethod
     def many_init(cls, *args, **kwargs):
         kwargs['child'] = cls(**kwargs)
-        try:
-            cls.Meta.depth = cls.compute_depth(kwargs['context']['view'].many_depth, cls.Meta.model, 'many_depth')
-        except KeyError:
-            pass
-        return ContainerSerializer(*args, **kwargs)
+        serializer = ContainerSerializer(*args, **kwargs)
+        if 'context' in kwargs and getattr(kwargs['context']['view'], 'nested_field', None) is not None:
+            serializer.id = '{}{}/'.format(serializer.id, kwargs['context']['view'].nested_field)
+        elif 'context' in kwargs:
+            serializer.id = '{}{}'.format(settings.BASE_URL, kwargs['context']['view'].request.path)
+        return serializer
+
+    def to_internal_value(self, data):
+        user_case = self.Meta.model is get_user_model() and '@id' in data and not data['@id'].startswith(
+            settings.BASE_URL)
+        if user_case:
+            data['username'] = 'external'
+        ret = super().to_internal_value(data)
+        if user_case:
+            ret['username'] = data['@id']
+        return ret
 
     def get_value(self, dictionary):
         try:
@@ -411,39 +527,48 @@ class LDPSerializer(HyperlinkedModelSerializer):
         info = model_meta.get_field_info(model_class)
         many_to_many = {}
         for field_name, relation_info in info.relations.items():
-            if relation_info.to_many and relation_info.reverse and not (
-                    field_name in validated_data) and not field_name is None:
+            if relation_info.to_many and relation_info.reverse and not field_name is None:
                 rel = getattr(instance._meta.model, field_name).rel
                 if rel.name in validated_data:
                     related = validated_data[rel.name]
                     getattr(instance, field_name).add(related)
 
     def internal_create(self, validated_data, model):
-
-        nested_fk_fields_name = list(filter(lambda key: isinstance(validated_data[key], dict), validated_data))
-        for field_name in nested_fk_fields_name:
-            field_dict = validated_data[field_name]
-            field_model = getattr(model, field_name).field.rel.model
-            slug_field = Model.slug_field(field_model)
-            if slug_field in field_dict:
-                kwargs = {slug_field: field_dict[slug_field]}
-                sub_inst = field_model.objects.get(**kwargs)
-            else:
-                sub_inst = self.internal_create(field_dict, field_model)
-            validated_data[field_name] = sub_inst
+        validated_data = self.resolve_fk_instances(model, validated_data)
 
         nested_fields = []
         nested_list_fields_name = list(filter(lambda key: isinstance(validated_data[key], list), validated_data))
         for field_name in nested_list_fields_name:
             nested_fields.append((field_name, validated_data.pop(field_name)))
 
+        info = model_meta.get_field_info(model)
+        many_to_many = []
+        for field_name, relation_info in info.relations.items():
+            if relation_info.to_many and relation_info.reverse and (
+                    field_name in validated_data) and not field_name is None:
+                many_to_many.append((field_name, validated_data.pop(field_name)))
+        validated_data = self.remove_empty_value(validated_data)
+        if model is get_user_model() and 'urlid' in validated_data and not 'username' in validated_data:
+            validated_data['username'] = validated_data.pop('urlid')
         instance = model.objects.create(**validated_data)
 
+        for field_name, value in many_to_many:
+            validated_data[field_name] = value
+
         self.save_or_update_nested_list(instance, nested_fields)
 
         return instance
 
+    def remove_empty_value(self, validated_data):
+        for attr, value in validated_data.items():
+            if value is '':
+                validated_data[attr] = None
+        return validated_data
+
     def update(self, instance, validated_data):
+        model = self.Meta.model
+        validated_data = self.resolve_fk_instances(model, validated_data)
+
         nested_fields = []
         nested_fields_name = list(filter(lambda key: isinstance(validated_data[key], list), validated_data))
         for field_name in nested_fields_name:
@@ -457,12 +582,47 @@ class LDPSerializer(HyperlinkedModelSerializer):
             else:
                 setattr(instance, attr, value)
 
+        self.save_or_update_nested_list(instance, nested_fields)
         instance.save()
 
-        self.save_or_update_nested_list(instance, nested_fields)
-
         return instance
 
+    def resolve_fk_instances(self, model, validated_data):
+        nested_fk_fields_name = list(filter(lambda key: isinstance(validated_data[key], dict), validated_data))
+        for field_name in nested_fk_fields_name:
+            field_dict = validated_data[field_name]
+            try:
+                field_model = getattr(model, field_name).field.rel.model
+            except:
+                # not fk
+                continue
+            slug_field = Model.slug_field(field_model)
+            sub_inst = None
+            if 'urlid' in field_dict:
+                # has urlid and is a local resource
+                if parse.urlparse(settings.BASE_URL).netloc == parse.urlparse(field_dict['urlid']).netloc:
+                    # try slug field if it exists
+                    if slug_field in field_dict:
+                        kwargs = {slug_field: field_dict[slug_field]}
+                        sub_inst = field_model.objects.get(**kwargs)
+                    else:
+                        model, sub_inst = Model.resolve(field_dict['urlid'])
+                # remote resource - get backlinked copy
+                elif hasattr(field_model, 'urlid'):
+                    kwargs = {'urlid': field_dict['urlid']}
+                    sub_inst = field_model.objects.get(**kwargs)
+                elif issubclass(field_model, AbstractUser):
+                    kwargs = {'username': field_dict['urlid']}
+                    sub_inst = field_model.objects.get(**kwargs)
+            # try slug field, assuming that this is a local resource
+            elif slug_field in field_dict:
+                kwargs = {slug_field: field_dict[slug_field]}
+                sub_inst = field_model.objects.get(**kwargs)
+            if sub_inst is None:
+                sub_inst = self.internal_create(field_dict, field_model)
+            validated_data[field_name] = sub_inst
+        return validated_data
+
     def update_dict_value(self, attr, instance, value):
         info = model_meta.get_field_info(instance)
         slug_field = Model.slug_field(instance)
@@ -477,8 +637,20 @@ class LDPSerializer(HyperlinkedModelSerializer):
         if relation_info.to_many:
             value = self.internal_create(validated_data=value, model=relation_info.related_model)
         else:
-            value[instance._meta.fields_map[attr].remote_field.name] = instance
-            oldObj = getattr(instance, attr, None)
+            try:
+                reverse_attr_name = instance._meta.fields_map[attr].remote_field.name
+                many = False
+            except:
+                rel = list(filter(lambda field: field.name == attr, instance._meta.fields))[0].rel
+                many = rel.one_to_many
+                reverse_attr_name = rel.related_name
+            if many:
+                value[reverse_attr_name] = [instance]
+                oldObj = rel.model.object.get(id=value['urlid'])
+            else:
+                value[reverse_attr_name] = instance
+                oldObj = getattr(instance, attr, None)
+
             if oldObj is None:
                 value = self.internal_create(validated_data=value, model=relation_info.related_model)
             else:
@@ -499,19 +671,20 @@ class LDPSerializer(HyperlinkedModelSerializer):
     def save_or_update_nested_list(self, instance, nested_fields):
         for (field_name, data) in nested_fields:
             manager = getattr(instance, field_name)
-            slug_field = Model.slug_field(instance)
+            field_model = manager.model
+            slug_field = Model.slug_field(manager.model)
             try:
                 item_pk_to_keep = list(map(lambda e: e[slug_field], filter(lambda x: slug_field in x, data)))
             except TypeError:
                 item_pk_to_keep = list(
                     map(lambda e: getattr(e, slug_field), filter(lambda x: hasattr(x, slug_field), data)))
 
-            for item in list(manager.all()):
-                if not str(getattr(item, slug_field)) in item_pk_to_keep:
-                    if getattr(manager, 'through', None) is None:
+            if getattr(manager, 'through', None) is None:
+                for item in list(manager.all()):
+                    if not str(getattr(item, slug_field)) in item_pk_to_keep:
                         item.delete()
-                    else:
-                        manager.remove(item)
+            else:
+                manager.clear()
 
             for item in data:
                 if not isinstance(item, dict):
@@ -519,8 +692,19 @@ class LDPSerializer(HyperlinkedModelSerializer):
                     saved_item = item
                 elif slug_field in item:
                     kwargs = {slug_field: item[slug_field]}
-                    old_obj = manager.model.objects.get(**kwargs)
-                    saved_item = self.update(instance=old_obj, validated_data=item)
+                    saved_item = self.get_or_create(field_model, item, kwargs)
+                elif 'urlid' in item and settings.BASE_URL in item['urlid']:
+                    model, old_obj = Model.resolve(item['urlid'])
+                    if old_obj is not None:
+                        saved_item = self.update(instance=old_obj, validated_data=item)
+                    else:
+                        saved_item = self.internal_create(validated_data=item, model=field_model)
+                elif 'urlid' in item and issubclass(field_model, AbstractUser):
+                    kwargs = {'username': item['urlid']}
+                    saved_item = self.get_or_create(field_model, item, kwargs)
+                elif 'urlid' in item:
+                    kwargs = {'urlid': item['urlid']}
+                    saved_item = self.get_or_create(field_model, item, kwargs)
                 else:
                     rel = getattr(instance._meta.model, field_name).rel
                     try:
@@ -532,4 +716,13 @@ class LDPSerializer(HyperlinkedModelSerializer):
                     saved_item = self.internal_create(validated_data=item, model=manager.model)
 
                 if getattr(manager, 'through', None) is not None and manager.through._meta.auto_created:
+                    manager.remove(saved_item)
                     manager.add(saved_item)
+
+    def get_or_create(self, field_model, item, kwargs):
+        try:
+            old_obj = field_model.objects.get(**kwargs)
+            saved_item = self.update(instance=old_obj, validated_data=item)
+        except field_model.DoesNotExist:
+            saved_item = self.internal_create(validated_data=item, model=field_model)
+        return saved_item
